/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)hunt.c	5.4 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

static	jmp_buf deadline;

char uucplock[BUFSIZ];

/* tipx */
char opened_tty[128];

/*
 * getremote: get device name from dv=A,B,C,.. field in remote.
 * tip coding.
 * USAGE
 * Let host and devices be given in remote by
 * host:\
 *     dv=/dev/ttya,/dev/ttyb,/dev/ttyc
 * then getremote("host") returns
 *     "/dev/ttya" at the first call,
 *     "/dev/ttyb" at the second call,
 *     "/dev/ttyc" at the third call,
 * Kenar says,
 *   I dislike to use getremote,
 *   because it is too tricky for the person who reads the code.
 *   This function was located in remote.c
 *   in spite of being unused out of this file.
 *   I moved here to clarity the irregularity in reading "hunt".
 * -Kenar-
 */
static char *getremote(host)
	char *host;
{
	register char *cp;
	static char *next;
	static int lookedup = 0;

	if (!lookedup) {
		if (host == NOSTR) {
			fprintf(stderr, "tip: no host specified\n");
			exit(3);
		}
		getremcap(host);
		next = DV;
		lookedup++;
	}
	/*
	 * We return a new device each time we're called (to allow
	 *   a rotary action to be simulated)
	 */
	if (next == NOSTR)
		return (NOSTR);
	if ((cp = index(next, ',')) == NULL) {
		DV = next;
		next = NOSTR;
	} else {
		*cp++ = '\0';
		DV = next;
		next = cp;
	}
	return (DV);
}


sigfunc_t dead()
{
	FD = -1; /* tip: deadfl = 1; -Kenar- */
	longjmp(deadline, 1);
}

/*
 * hunt: hunt device to be opened. 
 * Main parts are tip coding.
 * 1995/12 tip+ mod. by Kenar
 *
 * Tip+ new usage is,
 * return:
 *    device name that is hunted
 *    NULL: not found or opened.
 * Hunt now cleans locking file if failed.
 * EXAMPLE
 * Let host is given in remote as bellow.
 *    host:\
 *       dv=/dev/ttya,/dev/ttyb,/dev/ttyc
 * and let /dev/ttya is being used.
 * Then hunt("host") will return "/dev/ttyb".
 * -Kenar-
 */
char *hunt(name)
	char *name;
{
	register char *cp, *cf, *ct;
	sigfunc_t (*oldf)();

	oldf = signal(SIGALRM, dead);
	while (cp = getremote(name)) {
		/*
		 * The name of the lock is pretty important. We don't
		 * want to introduce the following security problems:
		 * 1. Namecollision in lock file between different devices
		 * 2. uucplock is used to decide whether or not to call
		 *    acucntrl. Thus, non-/dev devices should have a different
		 *    lock name than devices in /dev.
		 */
		if (*cp != '/')	/* device name must be absolute path */
			continue;
		while (*cp == '/') cp++;
		cp--;
		if (strncmp("/dev/", cp, 5) == 0) {
			cf = cp + 5;
			while (*cf == '/') cf++;
		} else
			cf = cp;
		ct = uucplock;
		while (*cf) {
			if (*cf == '/') {
				if (ct == uucplock
				    || (ct > uucplock && *(ct-1) != '_'))
					*ct++ = '_';
			} else
				*ct++ = *cf;
			cf++;
		}
		*ct = '\0';
		/*
		 * At this point we are guranteed that any devices outside
		 * /dev will have a lock name starting with underscore (_),
		 * and that any slashes (/) in the pathname have been changed
		 * to underscores, and that multiple slashes are only made
		 * into a single underscore. Thus, any way of expressing
		 * a pathname will lead to a unique lock file name.
		 */
		/* deadfl = 0; */
		if (uu_lock(uucplock) < 0)
			continue;
		/*
		 * Straight through call units, such as the BIZCOMP,
		 * VADIC and the DF, must indicate they're hardwired in
		 *  order to get an open file descriptor placed in FD.
		 * Otherwise, as for a DN-11, the open will have to
		 *  be done in the "open" routine.
		 */
		if (!HW)
			break;

		strcpy(opened_tty,cp); /* tipx */

		if (setjmp(deadline) == 0) {
			alarm(10);
			FD = open(cp, O_RDWR);
			/* What is the value FD, if timeout ?
			 * It is obscure in the original tip code. ( Buggy )
			 * I set it -1 in dead().
			 * -Kenar-
			 */
		}
		alarm(0);
		if (FD < 0) {
			perror(cp);
		}
		else {
			char buf[1];
			
			ioctl(FD, TIOCEXCL, 0);
			ioctl(FD, TIOCHPCL, 0);
			
			/* tip+ coding
			 * Make sure the device is really connecting by
			 * sending '\r' to modem and trying to receive it.
			 * Why should we do so?
			 * We have cases FD > 0 and not really connecting
			 * 1. Hardware wired.
			 * 2. NxFaxMonitor(Black&White co.) is running.
			 * And we have a troubling case
			 * when nothing is sent to modem:
			 * 1. NxFaxMonitor(Black&White co.) is running.
			 * -Kenar-
			 */
			alarm(2);
			write(FD,"\r",1);
			read(FD,buf,1);
			alarm(0); /* cancel it */
			
			signal(SIGALRM, oldf);
			return (cp);
		}
		(void)uu_unlock(uucplock);
	}
	signal(SIGALRM, oldf);
	return NULL;
}
