/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)tip.h	5.4 (Berkeley) 9/2/88
 */

/*
 * Copyright (c) 1995  Kenji Arisawa. All right reserved.
 * License to freely use and distribute this software is hereby granted 
 * by the author, subject to the condition that this copyright notice 
 * remains intact.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES.
 */
 

/*
 * tip - terminal interface program
 */

#include <sys/types.h>
#include <sys/file.h>

#include <sgtty.h>
#include <signal.h>
#include <stdio.h>
#include <pwd.h>
#include <ctype.h>
#include <setjmp.h>
#include <errno.h>
#include <netdb.h>
#include <strings.h>
#include <stdlib.h>

#define ERR (-1)
#ifndef TRUE
#define	TRUE	(1)
#endif
#ifndef FALSE
#define FALSE	(0)
#endif

#undef CTRL
#define	CTRL(c)	((c)&037)

#if	defined(SUNOS4) || defined(__ANSI__) || defined(sparc) || defined(NeXT)
typedef void sigfunc_t;
#else
typedef int sigfunc_t;
#endif




/*
 * Remote host attributes
 */
char *DV;			/* UNIX device(s) to open */
char *CM;			/* initial connection message */
char *CU;			/* call unit if making a phone call */
char *AT;			/* acu type */
char *PN;			/* phone number(s) */
char DU;			/* this host is dialed up */
char HW;			/* this device is hardwired, see hunt.c */

/*
 * Definition of indices into variable table so
 *  value(DEFINE) turns into a static address.
 */

#define BEAUTIFY	0
#define BAUDRATE	1
#define DIALTIMEOUT	2
#define EOFREAD		3
#define EOFWRITE	4
#define EOL		5
#define ESCAPE		6
#define EXCEPTIONS	7
#define FORCE		8
#define FRAMESIZE	9
#define HOST		10
#define LOG		11
#define PHONES		12
#define PROMPT		13
#define RAISE		14
#define RAISECHAR	15
#define RECORD		16
#define REMOTE		17
#define SCRIPT		18
#define TABEXPAND	19
#define VERBOSE		20
#define SHELL		21
#define HOME		22
#define ECHOCHECK	23
#define DISCONNECT	24
#define TAND		25
#define LDELAY		26
#define CDELAY		27
#define ETIMEOUT	28
#define RAWFTP		29
#define UNUSED1		30
#define	LOCALECHO	31
#define	PARITY		32
#define	SLIPADDR	33
#define SLIPDSTADDR	34
#define	SLIPMASK	35
#define	SLIPTYPE	36
#define SYSSCRIPT	37
#define	SETROUTE	38
#define LINESYNC    39
#define DTRHUP		40
/* added by Kenar */
#define LANGTYPE	41
#define XFER		42
#define KEYMAP		43



/*
 * String value table
 */
typedef
	struct {
		char	*v_name;	/* whose name is it */
		char	v_type;		/* for interpreting set's */
		char	v_access;	/* protection of touchy ones */
		char	*v_abbre;	/* possible abbreviation */
		char	*v_rname;	/* abbreviation used in /etc/remote */
		void	*v_value;	/* pointer to data */
	}
	value_t;

/* I hate CPU dependency and so replaced original codes... -Kenar-*/
/* used only value.c */
#define number(v)	(*(int*)(v))
#define boolean(v)	(*(int*)(v))
#define character(v)	(*(char*)(v))

#define value(v)	(vtable[v].v_value)
#define svalue(v)	((char*)vtable[v].v_value)
#define nvalue(v)	(*(int*)vtable[v].v_value)
#define bvalue(v)	(*(int*)vtable[v].v_value)
#define cvalue(v)	(*(char*)vtable[v].v_value)

#define vname(v)	vtable[v].v_name
#define vtype(v)	vtable[v].v_type
#define vaccess(v)	vtable[v].v_access
#define vabbre(v)	vtable[v].v_abbre
#define vrname(v)	vtable[v].v_rname

/* Use vdef/vndef to detect the string value is defined or not */
#define vdef(v)		(*(char*)vtable[v].v_value != '\0')
#define vndef(v)	(*(char*)vtable[v].v_value == '\0')
/* Use vundef to eliminate the definition of the string value */
#define vundef(v)	(char*)vtable[v].v_value = ""

#define STRING	01		/* string valued */
#define BOOL	02		/* true-false value */
#define NUMBER	04		/* numeric value */
#define CHAR	010		/* character value */
#define TMASK	017		/* type mask */

#define	READ	01		/* read access */
#define WRITE	02		/* write access to variable */
#define SECURE	04		/* Inhibit the use of environment in daemon exec. */
#define CHANGED	010		/* low bit is used to show modification */


/*
 * Definition of ACU line description
 */
typedef
	struct {
		char	*acu_name;
		int	(*acu_dialer)();
		int	(*acu_disconnect)();
		int	(*acu_abort)();
	}
	acu_t;

#define	equal(a, b)	(strcmp(a,b)==0)/* A nice function to string compare */


/*
 * Escape command table definitions --
 *   lookup in this table is performed when ``escapec'' is recognized
 *   at the begining of a line (as defined by the eolmarks variable).
*/

typedef
	struct {
		char	e_char;		/* char to match on */
		char	e_flags;	/* experimental, priviledged */
		char	*e_help;	/* help string */
		void 	(*e_func)();	/* command */
	}
	esctable_t;

#define NORM	00		/* normal protection, execute anyone */
#define EXP	01		/* experimental, mark it with a `*' on help */
#define PRIV	02		/* priviledged, root execute only */

extern int	vflag;		/* verbose during reading of .tiprc file */
extern value_t	vtable[];	/* variable table */

#ifndef ACULOG
#define logent(a, b, c, d)
#define loginit()
#endif

#define NOACU	((acu_t *)NULL)
#define NOSTR	((char *)NULL)
#define NOFILE	((FILE *)NULL)
#define NOPWD	((struct passwd *)0)


int	FD;			/* open file descriptor to remote host */

extern char	uucplock[];		/* name of lock file for uucp's */

extern char *hunt(char*);
extern	char *ctrl();
extern	char *ctime();
extern	long time();
extern	struct passwd *getpwuid();
extern	char *getlogin();
extern	char *vinterp();
extern char *expand(char *s);
extern	char *getenv();

char *lconnect();
int size(register char *s);



/*
 * Addition by Kenar
 */
#define PATHSIZ 256 /* maximum path size we support */
/* child control through signal */
#define CWAIT  "CWAIT"   /* wait */
#define CRTHON "CRTHON"  /* raw through on */
#define CRTHOFF "CRTHOFF" /* raw through off */
#define CSWEEPON "CSWEEP:1" /* sweep out received chars */
#define CSWEEPOFF "CSWEEP:0" /* receive normally */

#define makstr(x) ((char*)strcpy((char*)malloc(strlen(x)+1),x))
#define lett unsigned char

/* Pointers to com. port I/O function introduced by Kenar */
/* They are defined by set_crwfun() in crw.c  */
int (*cread)();
int (*cwrite)();
char *strapp(char*,char*);
char *skip(char *s);
char *skip(char*);

#define pread(buf,size)  (*cread)(buf,size)
#define pwrite(buf,n)	(*cwrite)(buf,n)

#define repeat do
#define until(x) while(!(x))
#define YES 1
#define NO 0
/* end of Kenar's code */