/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/*+:EDITS:*/
/*:05-08-1990-15:29-wht@n4hgf-rawthru */
/*:05-08-1990-15:28-wht@n4hgf-mv tipout.c tipout-orig.c */

#ifndef lint
static char sccsid[] = "@(#)tipout.c	5.2 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"
/*
 * tip
 *
 * lower fork of tip -- handles passive side
 *  reading from the remote host
 */

extern int	fildes[2];		/* file transfer synchronization channel */
extern int	repdes[2];		/* read process sychronization channel */
extern char	ccc;			/* synchronization character */
extern int rawthru;
static FILE	*fscript;		/* FILE for scripting */
static	jmp_buf sigbuf;
static csweep = 0;

/*
 * TIPOUT wait state routine --
 *   sent by TIPIN when it wants to posses the remote host
 */
sigfunc_t
intIOT()
{
	char line[256];
	int n;

	n = 0;
	do{
		read(fildes[0], &line[n], 1);
	} while(line[n++]);
	
	if(equal(line,CWAIT)){
		write(repdes[1],&ccc,1);
		read(fildes[0], &ccc,1);
		longjmp(sigbuf, 1);
	}
	if(equal(line,CRTHON)) rawthru = 1;
	else if(equal(line,CRTHOFF)) rawthru = 0;
	else if(equal(line,CSWEEPON)) csweep = 1;
	else if(equal(line,CSWEEPOFF)) csweep = 0;
}

/*
 *  intEMT:
 *  Accepts changes of tip variables over the pipe and acts accordingly.
 *	see also sigEMT in cmds.c
 *
 *  Kenar says,
 *    All changes of tip variables should be accepted through intEMT.
 *  1995/12: Completely rewritten by Kenar for tip+ support.
 *	-Kenar-
 */
sigfunc_t
intEMT()
{
	char line[256],*s;
	int n,id;

	n = 0;
	do{
		read(fildes[0], &line[n], 1);
	} while(line[n++]);
	
	/* DEBUG 
	printf("intEMT::%s\r\n",line);
	*/
	s = index(line,':');
	*s++ = '\0';
	id = atoi(line);
	/* Don't use sscanf. STRING type data may not be printable. */
	switch(vtype(id)&TMASK){
	case STRING:
		svalue(id) = makstr(s);
		break;
	case NUMBER:
		nvalue(id) = atoi(s);
		break;
	case BOOL:
		bvalue(id) = atoi(s);
		break;
	case CHAR:
		cvalue(id) = *s;
		break;
	}
	
	switch(id){
		case RECORD:/* We should permit the change of record in scripting */
			if(fscript){
				fclose(fscript);
				fscript = NULL;
			}
			if(bvalue(SCRIPT)){
				if(vndef(RECORD))
					puts("record file is not specified\r");
				else {
					fscript = fopen(svalue(RECORD), "a");
					if(!fscript)
						printf("can't create %s\r\n", svalue(RECORD));
				}
			}
			break;
		case SCRIPT:/* I am afraid of receiving "SCRIPT:TRUE" in scripting */
			if(bvalue(SCRIPT)){
				if(fscript) break;
				if(vndef(RECORD))
					puts("record file is not specified\r");
				else {
					fscript = fopen(svalue(RECORD), "a");
					if(!fscript) printf("can't create %s\r\n", svalue(RECORD));
				}
			}
			else{
				if(fscript){
					fclose(fscript);
					fscript = NULL;
				}
			}
			break;
		case PARITY: setparity(); break;
		case LANGTYPE: setpio(); break;
		default:;
	};
	longjmp(sigbuf, 1);
}

sigfunc_t
intTERM()
{
	if (bvalue(SCRIPT) && fscript != NULL)
		fclose(fscript);
	exit(0);
}


/* tipx */
extern int rawthru;


sigfunc_t
sigbus()
{
	kill(getppid(),SIGBUS);
	sleep(5); /* To be killed by the parent */
}

sigfunc_t
sigsegv()
{
	kill(getppid(),SIGSEGV);
	sleep(5); /* To be killed by the parent */
}



/*
 * ****TIPOUT   TIPOUT****
 */
/* tip+ */
#define	ALLSIGS	sigmask(SIGEMT)|sigmask(SIGTERM)|sigmask(SIGIOT)

/* tipx: ----
#define	ALLSIGS	sigmask(SIGEMT)|sigmask(SIGTERM)|sigmask(SIGIOT)|sigmask(SIGUSR1)|sigmask(SIGUSR2)
*/

/* tip: ----
#define	ALLSIGS	sigmask(SIGEMT)|sigmask(SIGTERM)|sigmask(SIGIOT)|sigmask(SIGSYS)
*/

/*
 *	tipout() read chars from modem and put them to stdout.
 *	- Kenar -
 */
tipout()
{
	lett buf[BUFSIZ];
	register lett *cp;
	register int cnt;
	extern int errno;
	int omask;

	if(setjmp(sigbuf) == 0){
		signal(SIGINT, SIG_IGN);
		signal(SIGQUIT, SIG_IGN);
		signal(SIGEMT, intEMT);		/* attention from TIPIN */
		signal(SIGTERM, intTERM);	/* time to go signal */
		signal(SIGIOT, intIOT);		/* scripting going on signal */
		signal(SIGHUP, intTERM);	/* for dial-ups */
		/* tip+ */
		signal(SIGBUS,sigbus);
		signal(SIGSEGV,sigsegv);
	}

	for(;;)
	{	cnt = pread(buf, BUFSIZ);
		/*	Here all chars in port buffer are read.		 */
		/*
		 *	If buffer is empty, pread wait at least one char is received.
		 *	-Kenar -
	     */
		if (cnt < 0)
		{	/* lost carrier */
			if(errno == EIO)
			{	sigblock(sigmask(SIGTERM));
				intTERM();
				/*NOT REACHED*/
			}
			else continue;
		}
		else if(csweep) continue;
		else if(cnt > 0)
		{	/* block signals while writing. -Kenar - */
			omask = sigblock(ALLSIGS);
			write(1,buf,cnt);
			if (bvalue(SCRIPT) && fscript != NULL)
			{	if (bvalue(BEAUTIFY))
				{	for (cp = buf; cp < buf + cnt; cp++)
						if (!(*cp < ' ') ||
						any(*cp,svalue(EXCEPTIONS)))
							putc(*cp, fscript);
				}
				else
					fwrite(buf, 1, cnt, fscript);
			}
			sigsetmask(omask);
		}
	};
}
