/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)value.c	5.2 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

#define MIDDLE	35

extern int pid;

static void vtoken();
static void vprint();
static int vlookup();
static int col = 0;


/*
 * vassign: assign value to variable and put CHANGE flag
 * original usage is changed by Kenar.
 * new usage:
 * EXAMPLES
 * vassign(BAUDRATE,"9600")
 * vassign(RAISECHAR,"\001")
 *
 * 1995/12 tip+ mod. by Kenar
 * -Kenar-
 */
static void vassign(id, v)
	int id;
	char *v;
{
	if (!(vaccess(id)&WRITE)) {
		printf("%s access denied\r\n",vname(id));
		return;
	}
	switch (vtype(id)&TMASK) {

	case STRING:
		if (equal(svalue(id), v)) return;
		svalue(id) = makstr(v);
		break;

	case NUMBER:
		if (nvalue(id) == atoi(v)) return;
		nvalue(id) = atoi(v);
		break;

	case BOOL:
		if (bvalue(id) == (character(v) != '!')) return;
		bvalue(id) = (character(v) != '!');
		break;

	case CHAR:
		if (cvalue(id) == character(v)) return;
		cvalue(id) = character(v);
		break;
	}
	vaccess(id) |= CHANGED;
}



/*
 * vcheck: remove CHANGE mark, do some actions for change,
 * and send message to child if he exist.
 * vcheck follows vlex.
 * Note that "set record=blah script" do work correctly,
 * because RECORD < SCRIPT.
 *
 * vcheck should not be executed in child process.
 *
 * -Kenar-
 */
void vcheck()
{	int n;
	char *s;

	for(n = 0; vname(n); n++){
		if(vaccess(n)&CHANGED){
			vaccess(n) &= ~CHANGED;
			/* and we have some exceptions */
			switch(n){
			case TANDEM: tandem(); break;
			case PARITY: setparity(); break;
			case LANGTYPE: setpio(); break;
			case LDELAY:
				if(nvalue(LDELAY) <0) nvalue(LDELAY) = 0;
				if(nvalue(LDELAY) > 999) nvalue(LDELAY) = 999;
				break;
			case CDELAY:
				if(nvalue(CDELAY) <0) nvalue(CDELAY) = 0;
				if(nvalue(CDELAY) > 999) nvalue(CDELAY) = 999;
				break;
			case KEYMAP:
				if(vdef(KEYMAP)){
					s = svalue(KEYMAP);
					if(strlen(s) != 3 || s[1] != ':'){
						puts("Format is X:Y where X,Y are characters.\r");
						vundef(KEYMAP);
					}
				}
				break;
			case RECORD:
				svalue(RECORD) = makstr(expand(s));
				break;
			default:;
			}
			if(pid){ /* Let child know the change. See tipout.c */
				sigEMT(n);
				nap(100); /* sleep 100ms */
			}
		}
	}
}


/*
 * vlex: set or print variables complying with a input of [set] command.
 * vlex is used in precedence to vcheck.
 * variable definition in tip+ script (system and/or users), in '.tiprc'
 * and variable setting in [set] command uses vlex.
 * EXAMPLES
 * vlex("all") prints all tip variables
 * vlex("?") prints variables of WRITE permission.
 * vlex("W? X=a Y=b !Z")
 * 
 * 1995/12	slight modification for readability by Kenar
 * -Kenar-
 */
void vlex(s) char *s;
{
	int n;;

	if(equal(s, "all")) {
		for (n = 0; vname(n); n++)
			if (vaccess(n)&READ) vprint(n);
	} else if(equal(s,"?")) {
		for (n = 0; vname(n); n++)
			if (vaccess(n)&WRITE) vprint(n);
	} else {
		char *cp;

		do {
			if (cp = vinterp(s, ' '))
				cp++;
			vtoken(s);
			s = cp;
		} while (s);
	}
	if (col > 0) {
		printf("\r\n");
		col = 0;
	}
}


/*
 * vtoken: set or print value of single variable.
 * EXAMPLE
 * vtoken("X=a")
 * vtoken("!Y")
 * vtoken("X?")
 *
 * 1995/12	slight modification for readablity by Kenar
 * -Kenar-
 */
static void
vtoken(s)
	char *s;
{
	int n;
	char *cp;

	if (cp = index(s, '=')) {
		*cp = '\0';
		if ((n = vlookup(s)) != ERR) {
			cp++;
			vassign(n, cp);
			return;
		}
	} else if (cp = index(s, '?')) {
		*cp = '\0';
		if ((n = vlookup(s)) != ERR && (vaccess(n)&READ)) {
			vprint(n);
			return;
		}
	} else {
		if (*s != '!')
			n = vlookup(s);
		else
			n = vlookup(s+1);
		if (n != ERR) {
			vassign(n, s);
			return;
		}
	}
	printf("%s: unknown variable\r\n", s);
}


/*
 * print value of tip variable.
 *
 * 1995/12	slight modification for readablity by Kenar
 * -Kenar-
 */
static void
vprint(n) int n;
{
	char *cp;
	extern char *interp(), *ctrl();

	if (col > 0 && col < MIDDLE)
		while (col++ < MIDDLE)
			putchar(' ');
	col += size(vname(n));
	switch (vtype(n)&TMASK) {

	case BOOL:
		if (bvalue(n) == FALSE) {
			col++;
			putchar('!');
		}
		printf("%s", vname(n));
		break;

	case STRING:
		printf("%s=", vname(n));
		col++;
		cp = interp(svalue(n), NULL);
		col += size(cp);
		printf("%s", cp);
		break;

	case NUMBER:
		col += 6;
		printf("%s=%-5d", vname(n), nvalue(n));
		break;

	case CHAR:
		printf("%s=", vname(n));
		col++;
		cp = ctrl(cvalue(n));
		col += size(cp);
		printf("%s", cp);
		break;
	}
	if (col >= MIDDLE) {
		col = 0;
		printf("\r\n");
		return;
	}
}


/*
 * look up tip variable using name.
 * return: variable id.
 * EXAMPLE
 * vlookup("ba") returns BAUDRATE
 *
 * 1995/12	slight modification for readablity by Kenar
 * -Kenar-
 */
int vlookup(s) char *s;
{
	int n;

	for (n = 0; vname(n); n++)
		if (equal(vname(n), s) || (vabbre(n) && equal(vabbre(n), s)))
			return (n);
	return (ERR);
}

char *
vinterp(s, stop)
	char *s;
	char stop;
{
	register char *p = s, c;
	int num;

	while ((c = *s++) && c != stop)
		switch (c) {

		case '^':
			if (*s)
				*p++ = *s++ - 0100;
			else
				*p++ = c;
			break;

		case '\\':
			num = 0;
			c = *s++;
			if (c >= '0' && c <= '7')
				num = (num<<3)+(c-'0');
			else {
				register char *q = "n\nr\rt\tb\bf\f";

				for (; *q; q++)
					if (c == *q++) {
						*p++ = *q;
						goto cont;
					}
				*p++ = c;
			cont:
				break;
			}
			if ((c = *s++) >= '0' && c <= '7') {
				num = (num<<3)+(c-'0');
				if ((c = *s++) >= '0' && c <= '7')
					num = (num<<3)+(c-'0');
				else
					s--;
			} else
				s--;
			*p++ = num;
			break;

		default:
			*p++ = c;
		}
	*p = '\0';
	return (c == stop ? s-1 : NULL);
}

